/*
 * Copyright 2017 The Netty Project
 *
 * The Netty Project licenses this file to you under the Apache License,
 * version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package io.netty.microbench.handler.ssl;

import io.netty.buffer.ByteBuf;
import org.openjdk.jmh.annotations.Benchmark;
import org.openjdk.jmh.annotations.Param;

public class SslHandlerEchoBenchmark extends AbstractSslHandlerThroughputBenchmark {
    @Param({ "1", "2", "5", "10" })
    public int numWrites;

    @Benchmark
    public ByteBuf wrapUnwrap() throws Exception {
        ByteBuf src = doWrite(numWrites);

        do {
            serverSslHandler.channelRead(serverCtx, src);
        } while (src.isReadable());

        assert !src.isReadable() && src.refCnt() == 1 : "src: " + src + " src.refCnt(): " + src.refCnt();

        return src;
    }
}
